package asd.td1;

/**
 * Une classe pour les suites d'entiers, pour les exercices 7  10 de la
 * feuille 1.
 * Une suite d'entiers (ou plus simplement, une suite) est un objet contenant
 * des entiers. On ne peut pas modifier une suite (c'est  dire ajouter ou 
 * retirer un entier de la suite). On accde aux lments de la suite via un
 * curseur.
 *  2002 Marc Gatano
 */
public class Suite {

    // la suite est implmente  l'aide d'un tableau et d'un indice
    // (le curseur)

    private int[] t;
    private int i;

    /**
     * Construit une suite contenant exactement les lments de table.
     * Complexit : O(1).
     */
    public Suite(int[] table) {
	t = table;
	i = 0;
    }

    /**
     * Construit la suite vide.
     * Complexit : O(1).
     */
    public Suite() {
	t = new int[0];
	i = 0;
    }

    /**
     * Place le curseur de la suite au dbut (c'est  dire sur le premier
     * lment ou en fin de suite si la suite est vide).
     * Complexit : O(1).
     */
    public void debut() {
	i = 0;
    }

    /**
     * Si le curseur n'est pas en fin de suite, retourne l'entier point
     * par le curseur.
     * Prcondition : fin() = false.
     * Complexit : O(1).
     */
    public int valeurCourante() {
	return t[i];
    }

    /**
     * Si le curseur n'est pas en fin de suite, dplace le curseur sur
     * l'lment suivant, ou en fin de suite si l'lment courant tait
     * le dernier.
     * Prcondition : fin() = false.
     * Complexit : O(1).
     */
    public void suivant() {
	i++;
    }

    /**
     * Indique si on est en fin de suite, autrement dit si le curseur pointe
     * au-del du dernier lment.
     * Complexit : O(1).
     */
    public boolean fin() {
	return i == t.length;
    }
    
    // pour afficher l'lment courant
    //
    private String toString(int j) {
	if ( j == i ) {
	    return "|" + String.valueOf(t[j]) + "|";
	}
	else {
	    return String.valueOf(t[j]);
	}
    }

    /**
     * La reprsentation externe d'une suite.
     * Complexit : THETA(n) o n est la longueur de la suite.
     */
    public String toString() {
	if ( t.length == 0 ) {
	    return "<>";
	}
	String s = "<" + toString(0);
	for ( int j = 1; j < t.length; j++ ) {
	    s = s + ", " + toString(j);
	}
	s = s + ">";
	return s;
    }

    // Une suite est reprsente par <AV,vC,AP> o <AV> est la sous-suite des
    // lments placs AVant la valeur courante (les lments lus), vC la
    // valeur courante et <AP> la sous-suite des lments placs APrs la
    // valeur courante (les lments  lire).

    // Exercice 7 /////////////////////////////////////////////////////////

    /**
     * Teste si la suite est croissante.
     * Complexit : O(n) o n est la longueur de la suite.
     */
    public boolean croissante() {
        debut();
        
        if ( fin() ) {
            return true;
	}
	
	int prec = valeurCourante();
        suivant();

        // INVARIANT : <AV> est croissante et prec est la dernire
        // valeur de <AV>
        // VARIANT : la longueur de <AV> crot et est major par
        // n o n est la longueur de la suite
        //
	while ( ! fin() && prec <= valeurCourante() ) {
	    prec = valeurCourante();
	    suivant();
	}

	return fin();
    }

    // Exercice 8 /////////////////////////////////////////////////////////

    /**
     * Retourne le nombre de monotonies de la suite.
     * Complexit : THETA(n) o n est la longueur de la suite.
     */
    public int monotonies() {
        debut();

        if ( fin() ) {
            return 0;
        }

        int prec = valeurCourante();
        suivant();

        // INVARIANT : <AV> est constante et prec est la dernire
        // valeur de <AV>
        // VARIANT : la longueur de <AV> crot et est major par
        // n o n est la longueur de la suite.
        //
	while ( ! fin() && prec == valeurCourante() ) {
	    prec = valeurCourante();
	    suivant();
	}

        if ( fin() ) {
            return 1;
        }

        boolean croissante = valeurCourante() > prec;
        int m = 1;
        suivant();

        // INVARIANT :
        // - il y a m monotonies dans <AV>
        // - prec est la dernire valeur de <AV>
        // - croissante <==> la dernire monotonie de <AV> est croissante
        // VARIANT : la longueur de <AV> crot et est major par
        // n o n est la longueur de la suite.
        //
        while ( ! fin() ) {
            if ( valeurCourante() < prec &&   croissante ||
                 valeurCourante() > prec && ! croissante ) {
                m++;
                croissante = ! croissante;
            }
            prec = valeurCourante();
            suivant();
        }
        return m;
    }

    // Exercice 9 /////////////////////////////////////////////////////////

    /**
     * Retourne la longueur d'une plus longue monotonie croissante
     * de la suite.
     * Complexit : THETA(n) o n est la longueur de la suite.
     */
    public int longueur() {
        debut();

        if ( fin() ) {
            return 0;
        }

        int prec = valeurCourante();
        suivant();

        int lmax = 0;
        int lder = 1;

        // INVARIANT :
        // - lmax est la longueur de la plus longue monotonie croissante
        //   sauf la dernire de <AV>
        // - lder est la longueur de la dernire monotonie croissante
        //   de <AV>
        //
        while ( ! fin() ) {
            if ( valeurCourante() < prec ) {
                if ( lder > lmax ) {
                    lmax = lder;
                }
                lder = 1;
            }
            else {
                lder++;
            }
            prec = valeurCourante();
            suivant();
        }

        if ( lder > lmax ) {
            return lder;
        }
        return lmax;
    }

    // Exercice 10 /////////////////////////////////////////////////////////

    /**
     * Avance le curseur sur la premire valeur diffrente de la valeur
     * courante.
     * Prcondition : fin() est faux.
     * Complexit : O(n) o n est la longueur de la suite.
     */
    private void suivantDifferent() {
        // p0 = valeurCourante()
	int p = valeurCourante();
	do {
	    suivant();
	} while ( ! fin() && p == valeurCourante() );
        //
        // INVARIANT : <AV> est constante et p est l'unique valeur de <AV>
        // VARIANT : la longueur de <AV> crot et est major par
        // n o n est la longueur de la suite.
    }

    /**
     * Teste si this et s sont semblables.
     * Complexit : O(n + m) o n et m sont respectivement
     * les longueurs de this et de s.
     */
    public boolean semblable(Suite s) {
        debut();
        s.debut();

        // on pose this = <AV,vC,AP> et s = <AV',vC',AP'>, et
        // p = la dernire valeur de <AV> et p' = la dernire valeur de <AV'>
        // INVARIANT :
        // - <AV> et <AV'> sont semblables
        // - p != vC et p' != vC'
        // VARIANT :
        // les longueurs respectives de <AV> et de <AV'> crossent et sont
        // majores par n et m  o n et m sont respectivement les longueurs
        // de la suite et de s.
        //
        while ( ! fin() && ! s.fin() &&
                valeurCourante() == s.valeurCourante() ) {

	    suivantDifferent();
	    s.suivantDifferent();
        }

        return fin() && s.fin();
    }
}
